/*
 * The contents of this file are subject to the terms 
 * of the Common Development and Distribution License 
 * (the License).  You may not use this file except in
 * compliance with the License.
 * 
 * You can obtain a copy of the license at 
 * https://glassfish.dev.java.net/public/CDDLv1.0.html or
 * glassfish/bootstrap/legal/CDDLv1.0.txt.
 * See the License for the specific language governing 
 * permissions and limitations under the License.
 * 
 * When distributing Covered Code, include this CDDL 
 * Header Notice in each file and include the License file 
 * at glassfish/bootstrap/legal/CDDLv1.0.txt.  
 * If applicable, add the following below the CDDL Header, 
 * with the fields enclosed by brackets [] replaced by
 * you own identifying information: 
 * "Portions Copyrighted [year] [name of copyright owner]"
 * 
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 */
package com.sun.enterprise.web.connector.grizzly.realtime;


import com.sun.enterprise.web.connector.grizzly.Pipeline;
import com.sun.enterprise.web.connector.grizzly.Task;
import com.sun.enterprise.web.connector.grizzly.WorkerThread;
import java.nio.ByteBuffer;

import javax.realtime.RealtimeThread;
import javax.realtime.PriorityParameters;
import javax.realtime.PriorityScheduler;

/**
 * Simple Real Time thread used for processing HTTP requests. All threads are
 * synchronized using a <code>Pipeline</code> object
 *
 * @author Jean-Francois Arcand
 */
public class RealTimeWorkerThread extends RealtimeThread implements WorkerThread{

    /**
     * The <code>ByteBuffer</code> used when <code>Task</code> are executed.
     */
    private ByteBuffer byteBuffer;
    
    
    /**
     * The <code>Pipeline</code> on which this thread synchronize.
     */
    private Pipeline pipeline;

    /**
     * Looing variable.
     */
    private volatile boolean doTask = true;

    
    /**
     * The <code>ThreadGroup</code> used.
     */
    private final static ThreadGroup threadGroup 
            = new ThreadGroup("GrizzlyRealTime");
    
    
    /** 
     * Create a Thread that will synchronizes/block on throws 
     * <code>Pipeline</code> instance.
     */
    public RealTimeWorkerThread(Pipeline pipeline){    
        this.pipeline = pipeline;
    }    
   
    
    /** 
     * Create a Thread that will synchronizes/block on throws 
     * <code>Pipeline</code> instance and that runs at a specified priority
     */
    public RealTimeWorkerThread(PriorityParameters pp, Pipeline pipeline){          
	super(pp);
        this.pipeline = pipeline;
    }   
    
    
    /**
     * Perform a <code>Task</code>.processing
     */
    public void run(){
    // This is a workaround for a bug in setting realtime priorities.
        PriorityScheduler scheduler = PriorityScheduler.instance();
        PriorityParameters pp = (PriorityParameters)this.getSchedulingParameters();
        pp.setPriority(scheduler.getMaxPriority()-10);
        while (doTask) {
            try{
                // Wait for a Task to be added to the pipeline.
                Task t = pipeline.getTask();
                if ( t != null){

                System.out.println("running task in RealtimeThread at priority " + 
				       pp);
                    t.run();                
                    t = null;
                }
            } catch (Throwable t) {
                t.printStackTrace();
            }
        }
    }
    
    
    /**
     * Stop this thread. If this Thread is performing atask, the task will be
     * completed.
     */
    public void terminate(){
        doTask = false;
    }
    
    
    /**
     * Set the <code>ByteBuffer</code> shared this thread
     */
    public void setByteBuffer(ByteBuffer byteBuffer){
        this.byteBuffer = byteBuffer;
    }
    
    
    /**
     * Return the <code>ByteBuffer</code> shared this thread
     */
    public ByteBuffer getByteBuffer(){
        return byteBuffer;
    }
    
    
}

